/***********************************************************/
//  BeUAE - The Be Un*x Amiga Emulator
//
//  BeOS port sound routines
//  Using R4 BSoundPlayer class
//
//  (c) 2000-2001 Axel Drfler
//  (c) 1999 Be/R4 Sound - Raphael Moll
//  (c) 1998-1999 David Sowsy
//
// History:
//
// -RM/050999 : removed debug code, code cleanup, etc.
// -DS/051399 : adapted code to build under 0.8.8.
//				New Media Kit woes ensued.	
// -DS/06xx99 : adapted code to work with the new Game Kit.
// -AD/121300 : supports now different frequencies, bitsrates, and stereo modes
// -AD/121500 : blocks now correctly to synchronize UAE
//
/***********************************************************/

#include "be-UAE.h"

#include <MediaDefs.h>

extern "C" {
#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "options.h"
#include "uae.h"
#include "gensound.h"
#include "sounddep/sound.h"
#include "events.h"
#include "custom.h"
};

extern "C" 
{
	extern int smplcnt;
	extern int init_sound(void);
	extern int setup_sound(void);
	extern void flush_sound_buffer(void);
//	extern void check_sound_buffers(void);
	extern void init_sound_table16(void);
	extern void sample16_handler(void);
};

uae_u16 *sndbuffer;
uae_u16 *sndbufpt;
int sndbufsize;
BSoundPlayer	*gSoundPlayer = NULL;
static int32	gSoundBufferSize = DEFAULT_SOUND_MINB;
static uae_u16	*gDoubleBufferRead = NULL;	// gSoundPlayer reads from this one
static uae_u16	*gDoubleBufferWrite = NULL;	// UAE writes to this one
static uae_u16	*gLastBuffer = NULL;		// the last buffer played
static int32	gBufferReadPos = 0;

static uae_u16 *buffer = NULL;
static bool sound_ready = false;
static sem_id sound_sync_sem;

void stream_func8(void *user, void *buffer, size_t size, const media_raw_audio_format &format);
void stream_func16(void *user, void *buffer, size_t size, const media_raw_audio_format &format);


int init_sound(void)
{ 
	if (gSoundPlayer != NULL)
		return 0;

	status_t err;
	BMediaRoster *gMediaRoster;
	media_node *outNode;
	outNode = new media_node;
	gMediaRoster = BMediaRoster::Roster(&err);
	
	if (gMediaRoster && err == B_OK)
		err = gMediaRoster->GetAudioOutput(outNode);

	if ((!gMediaRoster) || (err != B_OK))
	{ 
		printf("NO MEDIA ROSTER! The media server "
				"appears to be dead.\n"
				"\t-- roster %p -- error %08lx (%ld)\n",
				gMediaRoster, err, err);

		sound_ready = false;
		return 0;
	}
	
	media_raw_audio_format audioFormat;// = media_raw_audio_format::wildcard;
	audioFormat.frame_rate = currprefs.sound_freq;
	audioFormat.channel_count = currprefs.stereo ? 2 : 1;
	audioFormat.format = media_raw_audio_format::B_AUDIO_FLOAT;
	audioFormat.byte_order = B_MEDIA_HOST_ENDIAN;
	audioFormat.buffer_size = currprefs.sound_maxbsiz * sizeof(float);
printf("size = %ld\n",currprefs.sound_maxbsiz);
	gSoundPlayer = new BSoundPlayer(&audioFormat,"UAE SoundPlayer",currprefs.sound_bits == 8 ? stream_func8 : stream_func16);
	sound_ready = (gSoundPlayer != NULL);
    
	if (!currprefs.produce_sound)
		return 3;

	sound_sync_sem = create_sem(0,"UAE Sound Sync Semaphore");
	gSoundBufferSize = currprefs.sound_maxbsiz;
	gBufferReadPos = 0;
	gDoubleBufferWrite = new uae_u16[2 * gSoundBufferSize];
	gDoubleBufferRead = gDoubleBufferWrite + gSoundBufferSize;

	buffer = gDoubleBufferWrite;
	memset(buffer,0,4 * gSoundBufferSize);
	sndbufpt = sndbuffer = buffer;

	if (currprefs.sound_bits == 8)
	{
		sndbufsize = sizeof(uae_u8) * gSoundBufferSize;	
		sample_handler = currprefs.stereo ? sample8s_handler : sample8_handler;
		init_sound_table8();
	}
	else
	{
		sndbufsize = sizeof(uae_u16) * gSoundBufferSize;
		
		if (currprefs.stereo)
		{
			sample_handler = (currprefs.sound_interpol == 0 ? sample16s_handler
					: currprefs.sound_interpol == 1 ? sample16si_rh_handler
					: sample16si_crux_handler);
		}
		else
		{
			sample_handler = (currprefs.sound_interpol == 0 ? sample16_handler
					: currprefs.sound_interpol == 1 ? sample16i_rh_handler
					: sample16i_crux_handler);
		}
		init_sound_table16();
	}
    eventtab[sample_evtime].handler = sample_handler;
          
	sample_evtime = (long)maxhpos * maxvpos * 50 / currprefs.sound_freq;

	sound_available = 1;
	
	if (gSoundPlayer)
	{
		gSoundPlayer->Start();
		gSoundPlayer->SetHasData(true);
		return 1;
	}
	return 0;
}


int setup_sound(void)
{
	sound_available = 1;
	return 1;
}


void close_sound(void)
{
	if (sound_ready)
	{
		if (gSoundPlayer)
		{
			gSoundPlayer->Stop();
			delete gSoundPlayer;
			gSoundPlayer = NULL;
		}
		sound_ready = false;
	}

	delete_sem(sound_sync_sem);

	if (gDoubleBufferRead < gDoubleBufferWrite)
		delete[] gDoubleBufferRead;
	else
		delete[] gDoubleBufferWrite;
	gDoubleBufferRead = gDoubleBufferWrite = NULL;
}


void flush_sound_buffer(void)
{
	if (sound_ready && acquire_sem(sound_sync_sem) == B_OK)
	{
		uae_u16 *p = gDoubleBufferRead;		// swap buffers
		gDoubleBufferRead = gDoubleBufferWrite;
		buffer = gDoubleBufferWrite = p;
	}
	sndbufpt = sndbuffer = buffer;
}


void stream_func16(void *user, void *buffer, size_t size,const media_raw_audio_format &format)
{
	uae_u16 *buf;
	int32 max_read_sample, avail_sample;

	// since the BSoundPlayer supports only B_AUDIO_FLOAT, it's
	// very unlikely that this will ever happen:
	// if (format.format != media_raw_audio_format::B_AUDIO_FLOAT) return;

	float *dest = (float *)buffer;
	int32 dest_sample = (int32)(size/sizeof(float));
	float *enddest = dest+dest_sample;

	max_read_sample = gSoundBufferSize;
	if (dest_sample < max_read_sample) max_read_sample = dest_sample;

	buf = gDoubleBufferRead + gBufferReadPos;
	avail_sample = gSoundBufferSize - gBufferReadPos;
	if (avail_sample < max_read_sample) max_read_sample = avail_sample;
	if (max_read_sample)
		gBufferReadPos += max_read_sample;

	const float ratio = 1.f/32768.f;
	while(max_read_sample--)	// copy the buffer to the stream
	{
		int16 a = (int16)(*(buf++));
		*(dest++) = ((float)a) * ratio;
	}
	
	// the buffer is no longer needed, so lets release it.
	// if UAE is not fast enough to swap the buffers during play time, the same
	// buffer will be played again
	if (gBufferReadPos == gSoundBufferSize)
	{
		gBufferReadPos = 0;
		if (gLastBuffer != gDoubleBufferRead)
		{
			gLastBuffer = gDoubleBufferRead;
			release_sem(sound_sync_sem);
		}
	}
	
	while(dest < enddest)
		*(dest++) = 0.f;
}


void stream_func8(void *user, void *buffer, size_t size,const media_raw_audio_format &format)
{
	int32 max_read_sample, avail_sample;
	uae_u8 *buf;

	float *dest = (float *)buffer;
	int32 dest_sample = (int32)(size/sizeof(float));
	float *enddest = dest+dest_sample;

	max_read_sample = gSoundBufferSize;
	if (dest_sample < max_read_sample) max_read_sample = dest_sample;

	buf = (uae_u8 *)gDoubleBufferRead + gBufferReadPos;
	avail_sample = gSoundBufferSize - gBufferReadPos;
	if (avail_sample < max_read_sample) max_read_sample = avail_sample;
	if (max_read_sample)
		gBufferReadPos += max_read_sample;

	const float ratio = 1.f/128.f;
	while(max_read_sample--)
		*(dest++) = ((float)*(buf++)-128) * ratio;

	if (gBufferReadPos == gSoundBufferSize)
	{
		gBufferReadPos = 0;
		if (gLastBuffer != gDoubleBufferRead)
		{
			gLastBuffer = gDoubleBufferRead;
			release_sem(sound_sync_sem);
		}
	}

	while(dest < enddest)
		*(dest++) = 0.f;
}

